@@ -1,5 +1,5 @@
 /*  This file is part of "xtrace"
- *  Copyright (C) 2005 Bernhard R. Link
+ *  Copyright (C) 2005,2006 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2 of the License, or
@@ -16,6 +16,8 @@
  */
 #include <config.h>
 
+// to get strndup
+#define _GNU_SOURCE
 #include <assert.h>
 #include <values.h>
 #include <sys/types.h>
@@ -29,6 +31,8 @@
 
 #include "xtrace.h"
 
+static const bool print_counts = false;
+static const bool print_offsets = false;
 
 static inline unsigned int padded(unsigned int s) {
 	return (s+3)&(~3);
@@ -81,6 +85,8 @@ struct extension {
 	unsigned char numsubrequests;
 	const struct event *events;
 	unsigned char numevents;
+	const char **errors;
+	unsigned char numerrors;
 };
 
 struct usedextension {
@@ -91,6 +97,14 @@ struct usedextension {
 	unsigned char first_error;
 };
 
+void free_usedextensions(struct usedextension *e) {
+	while( e != NULL ) {
+		struct usedextension *h = e->next;
+		free(e);
+		e = h;
+	}
+}
+
 struct expectedreply {
 	struct expectedreply *next;
 	u_int64_t seq;
@@ -138,6 +152,8 @@ static const char *findConstant(const struct constant *constants, unsigned long
 };
 
 #define OFS_LATER ((size_t)-1)
+#define ROUND_32 ((size_t)-1)
+#define ROUND { ROUND_32, "", ft_LASTMARKER, NULL}
 
 struct parameter {
 	/* The offset within the event, request, reply or Struct this
@@ -157,12 +173,16 @@ struct parameter {
 		/* counts for following strings, lists, ...
 		 * value-mask for LISTofFormat */
 		ft_STORE8, ft_STORE16, ft_STORE32,
+		/* to be ft_GET later into the store register */
+		ft_PUSH8, ft_PUSH16, ft_PUSH32,
 		/* bitfields: multiple values are possible */
 		ft_BITMASK8, ft_BITMASK16, ft_BITMASK32,
 		/* Different forms of lists: */
 		/*	- boring ones */
-		ft_STRING8, ft_LISTofCARD32, 
+		ft_STRING8, ft_LISTofCARD32, ft_LISTofATOM,
 		ft_LISTofCARD8, ft_LISTofCARD16, 
+		ft_LISTofUINT8, ft_LISTofUINT16, 
+		ft_LISTofUINT32, 
 		/*	- one of the above depening on last FORMAT */
 		ft_LISTofFormat,
 		/*	- iterate of list description in constants field */
@@ -184,10 +204,16 @@ struct parameter {
 		ft_EVENT, 
 		/* jump to other parameter list if matches */
 		ft_IF8, 
+		/* jump to other parameter list if matches atom name */
+		ft_IFATOM, 
 		/* set end of last list manually, (for LISTofVarStruct) */
-		ft_LASTMARKER, 
+		ft_LASTMARKER,
+		/* a ft_CARD32 looking into the ATOM list */
+		ft_ATOM,
 		/* always big endian */
-		ft_BE32
+		ft_BE32,
+		/* get the #ofs value from the stack. (0 is the last pushed) */
+		ft_GET
 		} type;
 	const struct constant *constants;
 };
@@ -200,15 +226,22 @@ static size_t printSTRING8(u8 *buffer,size_t buflen,const struct parameter *p,si
 	if( buflen - ofs <= len )
 		len = buflen - ofs;
 
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
 	printf("%s='",p->name);
 	while( len > 0 ) {
 		if( nr == maxshownlistlen ) {
 			fputs("'...",stdout);
 		} else if( nr < maxshownlistlen ) {
-			if( getCARD8(ofs)< ' ' || getCARD8(ofs) > 'z' )
-				printf("\\%03hho",getCARD8(ofs));
+			unsigned char c = getCARD8(ofs);
+			if( c == '\n' ) {
+				putchar('\\');putchar('n');
+			} else if( c == '\t' ) {
+				putchar('\\');putchar('t');
+			} else if( (c >= ' ' && c <= '~' ) )
+				putchar(c);
 			else
-				putchar(getCARD8(ofs));
+				printf("\\%03hho", c);
 		}
 		ofs++;len--;nr++;
 	}
@@ -226,6 +259,8 @@ static size_t printLISTofCARD8(u8 *buffer,size_t buflen,const struct parameter *
 	if( buflen - ofs <= len )
 		len = buflen - ofs;
 
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
 	printf("%s=",p->name);
 	while( len > 0 ) {
 		const char *value;
@@ -259,6 +294,8 @@ static size_t printLISTofCARD16(struct connection *c,u8 *buffer,size_t buflen,co
 	if( (buflen - ofs)/2 <= len )
 		len = (buflen - ofs)/2;
 
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
 	printf("%s=",p->name);
 	while( len > 0 ) {
 		const char *value;
@@ -273,9 +310,9 @@ static size_t printLISTofCARD16(struct connection *c,u8 *buffer,size_t buflen,co
 			u16 = getCARD16(ofs);
 			value = findConstant(p->constants,u16);
 			if( value )
-				printf("%s(0x%hx)",value,u16);
+				printf("%s(0x%hx)",value,(unsigned short int)u16);
 			else
-				printf("0x%04hx",u16);
+				printf("0x%04hx",(unsigned short int)u16);
 		}
 		len--;ofs+=2;nr++;
 	}
@@ -292,6 +329,8 @@ static size_t printLISTofCARD32(struct connection *c,u8 *buffer,size_t buflen,co
 	if( (buflen - ofs)/4 <= len )
 		len = (buflen - ofs)/4;
 
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
 	printf("%s=",p->name);
 	while( len > 0 ) {
 		const char *value;
@@ -306,9 +345,151 @@ static size_t printLISTofCARD32(struct connection *c,u8 *buffer,size_t buflen,co
 			u32 = getCARD32(ofs);
 			value = findConstant(p->constants,u32);
 			if( value )
-				printf("%s(0x%x)",value,u32);
+				printf("%s(0x%x)",value,(unsigned int)u32);
 			else
-				printf("0x%08x",u32);
+				printf("0x%08x",(unsigned int)u32);
+		}
+		len--;ofs+=4;nr++;
+	}
+	putchar(';');
+	return ofs;
+}
+
+static size_t printLISTofATOM(struct connection *c,u8 *buffer,size_t buflen,const struct parameter *p,size_t len, size_t ofs){
+	bool notfirst = false;
+	size_t nr = 0;
+
+	if( buflen < ofs )
+		return ofs;
+	if( (buflen - ofs)/4 <= len )
+		len = (buflen - ofs)/4;
+
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
+	printf("%s=",p->name);
+	while( len > 0 ) {
+		const char *value;
+		u_int32_t u32;
+
+		if( nr == maxshownlistlen ) {
+			fputs(",...",stdout);
+		} else if( nr < maxshownlistlen ) {
+			if( notfirst )
+				putchar(',');
+			notfirst = true;
+			u32 = getCARD32(ofs);
+			value = findConstant(p->constants,u32);
+			if( value )
+				printf("%s(0x%x)",value,(unsigned int)u32);
+			else if( (value = getAtom(c,u32)) == NULL )
+				printf("0x%x",(unsigned int)u32);
+			else
+				printf("0x%x(\"%s\")",(unsigned int)u32,value);
+		}
+		len--;ofs+=4;nr++;
+	}
+	putchar(';');
+	return ofs;
+}
+
+static size_t printLISTofUINT8(u8 *buffer,size_t buflen,const struct parameter *p,size_t len, size_t ofs){
+	bool notfirst = false;
+	size_t nr = 0;
+
+	if( buflen < ofs )
+		return ofs;
+	if( buflen - ofs <= len )
+		len = buflen - ofs;
+
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
+	printf("%s=",p->name);
+	while( len > 0 ) {
+		const char *value;
+		unsigned char u8;
+
+		if( nr == maxshownlistlen ) {
+			fputs(",...",stdout);
+		} else if( nr < maxshownlistlen ) {
+			if( notfirst )
+				putchar(',');
+			notfirst = true;
+			u8 = getCARD8(ofs);
+			value = findConstant(p->constants,u8);
+			if( value )
+				printf("%s(%d)",value,(unsigned int)u8);
+			else
+				printf("%d",(unsigned int)u8);
+		}
+		len--;ofs++;nr++;
+	}
+	putchar(';');
+	return ofs;
+}
+
+static size_t printLISTofUINT16(struct connection *c,u8 *buffer,size_t buflen,const struct parameter *p,size_t len, size_t ofs){
+	bool notfirst = false;
+	size_t nr = 0;
+
+	if( buflen < ofs )
+		return ofs;
+	if( (buflen - ofs)/2 <= len )
+		len = (buflen - ofs)/2;
+
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
+	printf("%s=",p->name);
+	while( len > 0 ) {
+		const char *value;
+		u_int16_t u16;
+
+		if( nr == maxshownlistlen ) {
+			fputs(",...",stdout);
+		} else if( nr < maxshownlistlen ) {
+			if( notfirst )
+				putchar(',');
+			notfirst = true;
+			u16 = getCARD16(ofs);
+			value = findConstant(p->constants,u16);
+			if( value )
+				printf("%s(%d)",value,(unsigned int)u16);
+			else
+				printf("%d",(unsigned int)u16);
+		}
+		len--;ofs+=2;nr++;
+	}
+	putchar(';');
+	return ofs;
+}
+
+static size_t printLISTofUINT32(struct connection *c,u8 *buffer,size_t buflen,const struct parameter *p,size_t len, size_t ofs){
+	bool notfirst = false;
+	size_t nr = 0;
+
+	if( buflen < ofs )
+		return ofs;
+	if( (buflen - ofs)/4 <= len )
+		len = (buflen - ofs)/4;
+
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
+	printf("%s=",p->name);
+	while( len > 0 ) {
+		const char *value;
+		u_int32_t u32;
+
+		if( nr == maxshownlistlen ) {
+			fputs(",...",stdout);
+		} else if( nr < maxshownlistlen ) {
+			if( notfirst )
+				putchar(',');
+			notfirst = true;
+			u32 = getCARD32(ofs);
+			value = findConstant(p->constants,u32);
+			if( value )
+				printf("%s(%x)",value,(unsigned int)u32);
+			else
+				printf("%x",(unsigned int)u32);
 		}
 		len--;ofs+=4;nr++;
 	}
@@ -328,15 +509,19 @@ struct value {
 static size_t printLISTofVALUE(struct connection *c,u8 *buffer,size_t buflen,const struct parameter *param,unsigned long valuemask, size_t ofs){
 
 	const struct value *v = (const struct value*)param->constants;
+	const char *atom;
 	bool notfirst = false;
 
 	assert( v != NULL );
 
 	if( ofs > buflen )
 		return ofs;
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
 	printf("%s={",param->name);
 	while( buflen > ofs && buflen-ofs >= 4 ) {
 		u_int32_t u32; u_int16_t u16; u_int8_t u8;
+		int32_t i32; int16_t i16; int8_t i8;
 		const char *constant;
 
 		if( v->name == NULL ) /* EOV */
@@ -349,17 +534,17 @@ static size_t printLISTofVALUE(struct connection *c,u8 *buffer,size_t buflen,con
 			putchar(' ');
 		notfirst = true;
 		/* this is funny, but that is the protocol... */
-		u32 = getCARD32(ofs);
-		u16 = u32 & 65535;
-		u8 = u32 & 255;
+		u32 = getCARD32(ofs); i32 = u32;
+		u16 = u32 & 65535; i16 = u16;
+		u8 = u32 & 255; i8 = u8;
 		if( v->type >= ft_BITMASK8 ) {
 			assert(v->type <= ft_BITMASK32 );
 			print_bitfield(v->name,v->constants,u32);
 			ofs += 4;v++;
 			continue;
 		}
-		assert( v->type < ft_STORE8 );
-		switch( v->type % 3 ) {
+		assert( v->type < ft_STORE8 || v->type == ft_ATOM );
+		switch( (v->type==ft_ATOM)?2:(v->type % 3) ) {
 		 case 0:
 			constant = findConstant(v->constants,u8);
 			break;
@@ -377,40 +562,46 @@ static size_t printLISTofVALUE(struct connection *c,u8 *buffer,size_t buflen,con
 		}
 		switch( v->type ) {
 		 case ft_INT8:
-			 printf("%hhd",u8);
+			 printf("%d",(int)i8);
 			 break;
 		 case ft_INT16:
-			 printf("%hd",u16);
+			 printf("%d",(int)i16);
 			 break;
 		 case ft_INT32:
-			 printf("%d",u32);
+			 printf("%d",(int)i32);
 			 break;
 		 case ft_UINT8:
-			 printf("%hhu",u8);
+			 printf("%u",(unsigned int)u8);
 			 break;
 		 case ft_UINT16:
-			 printf("%hu",u16);
+			 printf("%u",(unsigned int)u16);
 			 break;
 		 case ft_UINT32:
-			 printf("%u",u32);
+			 printf("%u",(unsigned int)u32);
 			 break;
 		 case ft_ENUM8:
 			 if( constant == NULL )
 				 fputs("unknown:",stdout);
 		 case ft_CARD8:
-			 printf("0x%02hhx",u8);
+			 printf("0x%02x",(unsigned int)u8);
 			 break;
 		 case ft_ENUM16:
 			 if( constant == NULL )
 				 fputs("unknown:",stdout);
 		 case ft_CARD16:
-			 printf("0x%04hx",u16);
+			 printf("0x%04x",(unsigned int)u16);
+			 break;
+		 case ft_ATOM:
+			 printf("0x%x",(unsigned int)u32);
+			 atom = getAtom(c, u32);
+			 if( atom != NULL )
+				 printf("(\"%s\")", atom);
 			 break;
 		 case ft_ENUM32:
 			 if( constant == NULL )
 				 fputs("unknown:",stdout);
 		 case ft_CARD32:
-			 printf("0x%08x",u32);
+			 printf("0x%08x",(unsigned int)u32);
 			 break;
 		 default:
 			 assert(0);
@@ -425,11 +616,29 @@ static size_t printLISTofVALUE(struct connection *c,u8 *buffer,size_t buflen,con
 	return ofs;
 }
 
-static size_t print_parameters(struct connection *c,const unsigned char *buffer,unsigned int len, const struct parameter *parameters, bool bigrequest);
+struct stack {
+	unsigned long *base;
+	int num;
+	int ofs;
+};
 
-static size_t printLISTofStruct(struct connection *c,u8 *buffer,size_t buflen,const struct parameter *p,size_t count, size_t ofs){
+static unsigned long getFromStack(struct stack *stack, size_t depth) {
+	assert(stack != NULL && stack->ofs > depth );
+	return stack->base[stack->ofs - 1 - depth];
+}
+
+static void push(struct stack *stack, unsigned long value) {
+	stack->base[stack->ofs] = value;
+	stack->ofs++;
+	assert(stack->ofs<stack->num);
+}
+static void pop(struct stack *stack UNUSED, struct stack *oldstack UNUSED) {
+}
+
+static size_t print_parameters(struct connection *c,const unsigned char *buffer,unsigned int len, const struct parameter *parameters, bool bigrequest, struct stack *oldstack);
+
+static size_t printLISTofStruct(struct connection *c,u8 *buffer,size_t buflen,const struct parameter *p,size_t count, size_t ofs, struct stack *stack){
 	bool notfirst = false;
-//	size_t ofs = (p->offset<0)?lastofs:p->offset;
 	/* This is a gross hack: the constants for ft_LISTofStruct are
 	 * in reality a parameter structure */
 	const struct parameter *substruct = (const void*)p->constants;
@@ -441,6 +650,8 @@ static size_t printLISTofStruct(struct connection *c,u8 *buffer,size_t buflen,co
 	len = substruct->offse;
 	substruct++;
 
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
 	printf("%s=",p->name);
 	while( buflen > ofs && buflen-ofs >= len && count > 0) {
 
@@ -455,7 +666,7 @@ static size_t printLISTofStruct(struct connection *c,u8 *buffer,size_t buflen,co
 			notfirst = true;
 			putchar('{');
 
-			print_parameters(c,buffer+ofs,len,substruct,false);
+			print_parameters(c,buffer+ofs,len,substruct,false,stack);
 
 			putchar('}');
 		}
@@ -464,7 +675,7 @@ static size_t printLISTofStruct(struct connection *c,u8 *buffer,size_t buflen,co
 	putchar(';');
 	return ofs;
 }
-static size_t printLISTofVarStruct(struct connection *c,u8 *buffer,size_t buflen,const struct parameter *p,size_t count, size_t ofs){
+static size_t printLISTofVarStruct(struct connection *c,u8 *buffer,size_t buflen,const struct parameter *p,size_t count, size_t ofs, struct stack *stack){
 	bool notfirst = false;
 //	size_t ofs = (p->offset<0)?lastofs:p->offset;
 	const struct parameter *substruct = (const void*)p->constants;
@@ -476,6 +687,8 @@ static size_t printLISTofVarStruct(struct connection *c,u8 *buffer,size_t buflen
 	len = substruct->offse;
 	substruct++;
 
+	if( print_offsets )
+		printf("[%d]",(int)ofs);
 	printf("%s=",p->name);
 	while( buflen > ofs && buflen-ofs >= len && count > 0) {
 		size_t lentoadd;
@@ -491,7 +704,7 @@ static size_t printLISTofVarStruct(struct connection *c,u8 *buffer,size_t buflen
 		notfirst = true;
 		putchar('{');
 
-		lentoadd = print_parameters(c,buffer+ofs,len,substruct,false);
+		lentoadd = print_parameters(c,buffer+ofs,buflen-ofs,substruct,false,stack);
 
 		putchar('}');
 		ofs += lentoadd; count--; nr++;
@@ -503,14 +716,16 @@ static size_t printLISTofVarStruct(struct connection *c,u8 *buffer,size_t buflen
 /* buffer must have at least 32 valid bytes */
 static void print_event(struct connection *c,const unsigned char *buffer);
 
-static size_t print_parameters(struct connection *c,const unsigned char *buffer,unsigned int len, const struct parameter *parameters,bool bigrequest) {
+static size_t print_parameters(struct connection *c,const unsigned char *buffer,unsigned int len, const struct parameter *parameters,bool bigrequest, struct stack *oldstack) {
 	const struct parameter *p;
 	unsigned long stored = INT_MAX;
 	unsigned char format = 0;
 	bool notfirst = false;
 	size_t lastofs = 0;
+	struct stack newstack = *oldstack;
 
 	for( p = parameters; p->name != NULL; p++ ) {
+		int8_t i8; int16_t i16; int32_t i32;
 #ifdef STUPIDCC
 		u_int8_t u8=0; u_int16_t u16=0; u_int32_t u32=0;
 		unsigned long l=0;
@@ -520,6 +735,7 @@ static size_t print_parameters(struct connection *c,const unsigned char *buffer,
 #endif
 		size_t ofs;
 		const char *value;
+		const char *atom;
 
 		if( p->offse == OFS_LATER )
 			ofs = lastofs;
@@ -537,14 +753,26 @@ static size_t print_parameters(struct connection *c,const unsigned char *buffer,
 			if( ofs < len && 
 			  /* some more overloading: */
 			  getCARD8(ofs) == (unsigned char)(p->name[0]) )
-				p = (struct parameter *)p->constants;
-			else
+				p = ((struct parameter *)p->constants)-1;
+			continue;
+		} else if( p->type == ft_IFATOM ) {
+			const char *atom;
+			if( ofs+4 >= len )
 				continue;
+			atom = getAtom(c, getCARD32(ofs));
+			if( atom == NULL )
+				continue;
+			if( strcmp(atom, p->name) == 0 )
+				p = ((struct parameter *)p->constants)-1;
+			continue;
 		}
 
 		switch( p->type ) {
 		 case ft_LASTMARKER:
-			 lastofs = ofs;
+			 if( ofs == ROUND_32 )
+				 lastofs = (lastofs+3)& ~3;
+			 else
+				 lastofs = ofs;
 			 continue;
 		 case ft_FORMAT8:
 			 if( ofs < len )
@@ -562,6 +790,18 @@ static size_t print_parameters(struct connection *c,const unsigned char *buffer,
 		 case ft_LISTofCARD32:
 			lastofs = printLISTofCARD32(c,buffer,len,p,stored,ofs);
 			continue;
+		 case ft_LISTofATOM:
+			lastofs = printLISTofATOM(c,buffer,len,p,stored,ofs);
+			continue;
+		 case ft_LISTofUINT8:
+			lastofs = printLISTofUINT8(buffer,len,p,stored,ofs);
+			continue;
+		 case ft_LISTofUINT16:
+			lastofs = printLISTofUINT16(c,buffer,len,p,stored,ofs);
+			continue;
+		 case ft_LISTofUINT32:
+			lastofs = printLISTofUINT32(c,buffer,len,p,stored,ofs);
+			continue;
 		 case ft_LISTofFormat:
 			switch( format ) {
 			 case 8:
@@ -579,13 +819,13 @@ static size_t print_parameters(struct connection *c,const unsigned char *buffer,
 			}
 			continue;
 		 case ft_Struct:
-			printLISTofStruct(c,buffer,len,p,1,ofs);
+			printLISTofStruct(c,buffer,len,p,1,ofs,&newstack);
 			continue;
 		 case ft_LISTofStruct:
-			lastofs = printLISTofStruct(c,buffer,len,p,stored,ofs);
+			lastofs = printLISTofStruct(c,buffer,len,p,stored,ofs,&newstack);
 			continue;
 		 case ft_LISTofVarStruct:
-			lastofs = printLISTofVarStruct(c,buffer,len,p,stored,ofs);
+			lastofs = printLISTofVarStruct(c,buffer,len,p,stored,ofs,&newstack);
 			continue;
 		 case ft_LISTofVALUE:
 			lastofs = printLISTofVALUE(c,buffer,len,p,stored,ofs);
@@ -594,12 +834,31 @@ static size_t print_parameters(struct connection *c,const unsigned char *buffer,
 			if( len >= ofs + 32 )
 				print_event(c,buffer+ofs);
 			continue;
+		 case ft_ATOM:
+			if( ofs + 4 > len )
+				continue;
+			if( print_offsets )
+				printf("[%d]",(int)ofs);
+			fputs(p->name,stdout);putchar('=');
+			u32 = getCARD32(ofs);
+			value = findConstant(p->constants,u32);
+			atom = getAtom(c, u32);
+			if( value != NULL )
+				printf("%s(0x%x)",value, (unsigned int)u32);
+			else if( atom == NULL )
+				printf("0x%x(unrecognized atom)",(unsigned int)u32);
+			else
+				printf("0x%x(\"%s\")",(unsigned int)u32, atom);
+			continue;
 		 case ft_BE32:
 			if( ofs + 4 > len )
 				continue;
 			fputs(p->name,stdout);putchar('=');
 			printf("0x%08x",(unsigned int)getBE32(ofs));
 			continue;
+		 case ft_GET:
+			stored = getFromStack(&newstack,p->offse);
+			continue;
 		 default:
 			break;
 		}
@@ -627,12 +886,20 @@ static size_t print_parameters(struct connection *c,const unsigned char *buffer,
 			print_bitfield(p->name,p->constants,l);
 			continue;
 		}
-		if( p->type >= ft_STORE8 ) {
+		if( p->type >= ft_PUSH8 ) {
+			assert(p->type <= ft_PUSH32 );
+			push(&newstack,l);
+			if( !print_counts)
+				continue;
+		} else if( p->type >= ft_STORE8 ) {
 			assert(p->type <= ft_STORE32);
 			stored = l;
-			continue;
+			if( !print_counts)
+				continue;
 		}
 		value = findConstant(p->constants,l);
+		if( print_offsets )
+			printf("[%d]",(int)ofs);
 		fputs(p->name,stdout);putchar('=');
 		if( value != NULL ) {
 			fputs(value,stdout);
@@ -640,44 +907,50 @@ static size_t print_parameters(struct connection *c,const unsigned char *buffer,
 		}
 		switch( p->type ) {
 		 case ft_INT8:
-			 printf("%hhd",u8);
+			 i8 = u8;
+			 printf("%d",(int)i8);
 			 break;
 		 case ft_INT16:
-			 printf("%hd",u16);
+			 i16 = u16;
+			 printf("%d",(int)i16);
 			 break;
 		 case ft_INT32:
-			 printf("%d",u32);
+			 i32 = u32;
+			 printf("%d",(int)i32);
 			 break;
+		 case ft_PUSH8:
+		 case ft_STORE8:
 		 case ft_UINT8:
-			 printf("%hhu",u8);
+			 printf("%u",(unsigned int)u8);
 			 break;
+		 case ft_PUSH16:
+		 case ft_STORE16:
 		 case ft_UINT16:
-			 printf("%hu",u16);
+			 printf("%u",(unsigned int)u16);
 			 break;
+		 case ft_PUSH32:
+		 case ft_STORE32:
 		 case ft_UINT32:
-			 printf("%u",u32);
+			 printf("%u",(unsigned int)u32);
 			 break;
 		 case ft_ENUM8:
 			 if( value == NULL )
 				 fputs("unknown:",stdout);
 		 case ft_CARD8:
-			 printf("0x%02hhx",u8);
+			 printf("0x%02x",(unsigned int)u8);
 			 break;
 		 case ft_ENUM16:
 			 if( value == NULL )
 				 fputs("unknown:",stdout);
 		 case ft_CARD16:
-			 printf("0x%04hx",u16);
+			 printf("0x%04x",(unsigned int)u16);
 			 break;
 		 case ft_ENUM32:
 			 if( value == NULL )
 				 fputs("unknown:",stdout);
 		 case ft_CARD32:
-			 printf("0x%08x",u32);
+			 printf("0x%08x",(unsigned int)u32);
 			 break;
-		 case ft_STORE8:
-		 case ft_STORE16:
-		 case ft_STORE32:
 		 case ft_BITMASK8:
 		 case ft_BITMASK16:
 		 case ft_BITMASK32:
@@ -686,14 +959,21 @@ static size_t print_parameters(struct connection *c,const unsigned char *buffer,
 		 case ft_LISTofCARD8:
 		 case ft_LISTofCARD16:
 		 case ft_LISTofCARD32:
+		 case ft_LISTofATOM:
+		 case ft_LISTofUINT8:
+		 case ft_LISTofUINT16:
+		 case ft_LISTofUINT32:
 		 case ft_LISTofFormat:
 		 case ft_LISTofVALUE:
 		 case ft_Struct:
 		 case ft_LISTofStruct:
 		 case ft_LISTofVarStruct:
 		 case ft_IF8:
+		 case ft_IFATOM:
 		 case ft_BE32:
+		 case ft_ATOM:
 		 case ft_LASTMARKER:
+		 case ft_GET:
 		 case ft_EVENT:
 			 assert(0);
 		}
@@ -701,6 +981,7 @@ static size_t print_parameters(struct connection *c,const unsigned char *buffer,
 			putchar(')');
 		}
 	}
+	pop(&newstack,oldstack);
 	return lastofs;
 }
 
@@ -732,6 +1013,21 @@ static bool requestQueryExtension(struct connection *c, bool pre, bool bigreques
 	return false;
 }
 
+static bool requestInternAtom(struct connection *c, bool pre, bool bigrequest UNUSED, struct expectedreply *reply) {
+	u_int16_t len;
+	if( pre )
+		return false;
+	if( reply == NULL)
+		return false;
+	if( c->clientignore <= 8 )
+		return false;
+	len = clientCARD16(4);
+	if( c->clientignore < (unsigned int)8 + len)
+		return false;
+	reply->data = newAtom(c->clientbuffer+8,len);
+	return false;
+}
+
 /* Reactions to some replies */
 
 static void replyListFontsWithInfo(struct connection *c,bool *ignore,bool *dontremove,void *data UNUSED) {
@@ -762,6 +1058,13 @@ static void replyQueryExtension(struct connection *c,bool *ignore UNUSED,bool *d
 	}
 }
 
+static void replyInternAtom(struct connection *c,bool *ignore UNUSED,bool *dontremove UNUSED,void *data) {
+	u_int32_t atom;
+	if( data == NULL )
+		return;
+	atom = serverCARD32(8);
+	internAtom(c, atom, data);
+}
 
 #define ft_COUNT8 ft_STORE8
 #define ft_COUNT16 ft_STORE16
@@ -798,8 +1101,13 @@ static inline void print_client_request(struct connection *c,bool bigrequest) {
 	unsigned char req = clientCARD8(0);
 	const struct request *r;
 	const char *extensionname = "";
-	size_t len;
 	bool ignore;
+	size_t len;
+	unsigned long stackvalues[30];
+	struct stack stack;
+	stack.base = stackvalues;
+	stack.num = 30;
+	stack.ofs = 0;
 
 	len = c->clientignore;
 	if( len > c->clientcount )
@@ -822,7 +1130,10 @@ static inline void print_client_request(struct connection *c,bool bigrequest) {
 				extensionname,req,
 				r->name
 		      );
-		print_parameters(c,c->clientbuffer,len,r->parameters, bigrequest);
+		if( r->parameters != NULL )
+			print_parameters(c,c->clientbuffer,len,r->parameters, bigrequest, &stack);
+		else
+			puts("obsolete without parameter description");
 		if( r->request_func != NULL )
 			(void)r->request_func(c,false,bigrequest,NULL);
 		putchar('\n');
@@ -853,6 +1164,11 @@ static inline void print_server_reply(struct connection *c) {
 	unsigned int cmd,seq;
 	struct expectedreply *replyto,**lastp;
 	size_t len;
+	unsigned long stackvalues[30];
+	struct stack stack;
+	stack.base = stackvalues;
+	stack.num = 30;
+	stack.ofs = 0;
 
 	len = c->serverignore;
 	if( len > c->servercount )
@@ -872,7 +1188,7 @@ static inline void print_server_reply(struct connection *c) {
 			if( !ignore ) {
 				printf("%03d:>:0x%04x:%u: Reply to %s: ", c->id, seq, (unsigned int)c->serverignore,replyto->from->name);
 				print_parameters(c,
-					c->serverbuffer,len,replyto->from->answers,false);
+					c->serverbuffer,len,replyto->from->answers,false,&stack);
 				putchar('\n');
 			}
 			if( !dontremove ) {
@@ -900,11 +1216,30 @@ const char *errors[] = {
 
 static inline void print_server_error(struct connection *c) {
 	unsigned int cmd = serverCARD8(1);
+	struct usedextension *u;
+	const char *errorname;
+
+	if( cmd < NUM(errors) )
+		errorname = errors[cmd];
+	else {
+		errorname = "unknown";
+		for( u = c->usedextensions; u != NULL ; u = u->next ) {
+			unsigned int i;
+			if( cmd < u->first_error )
+				continue;
+			i = cmd-u->first_error;
+			if( i >= u->extension->numerrors )
+				continue;
+			errorname = u->extension->errors[i];
+			break;
+		}
+
+	}
 	printf("%03d:>:%u:Error %hhu=%s: major=%u, minor=%u, bad=%u\n",
 			c->id,
 			(int)serverCARD16(2),
 			cmd,
-			(cmd<NUM(errors))?errors[cmd]:"unknown",
+			errorname,
 			(int)serverCARD16(8),
 			(int)serverCARD8(10),
 			(int)serverCARD32(4));
@@ -964,7 +1299,7 @@ void parse_client(struct connection *c) {
 		 if( c->clientcount == sizeof(c->clientbuffer) ) 
 			 printf("%03d:<: Warning: buffer filled!\n",c->id);
 		 else if( c->clientcount < l ) {
-			 printf("%03d:<: Warning: Waiting for rest of package (yet got %u of %u)!\n",c->id,c->clientcount,l);
+			 printf("%03d:<: Warning: Waiting for rest of package (yet got %u of %u)!\n",c->id,c->clientcount,(unsigned int)l);
 			 return;
 		 }
 		 c->clientignore = l;
@@ -1045,6 +1380,11 @@ void parse_server(struct connection *c) {
 static void print_event(struct connection *c,const unsigned char *buffer) {
 	const struct event *event;
 	u_int8_t code = getCARD8(0);
+	unsigned long stackvalues[30];
+	struct stack stack;
+	stack.base = stackvalues;
+	stack.num = 30;
+	stack.ofs = 0;
 
 	if( (code & 0x80) != 0 ) 
 		fputs("(generated) ",stdout);
@@ -1067,16 +1407,23 @@ static void print_event(struct connection *c,const unsigned char *buffer) {
 	} else
 		event = &events[code];
 	printf("%s(%hhu) ",event->name,code);
-	print_parameters(c,buffer,32,event->parameters,false);
+	print_parameters(c,buffer,32,event->parameters,false,&stack);
 }
 
 #include "shape.inc"
 #include "bigrequest.inc"
+#include "render.inc"
+#include "randr.inc"
+#include "mitshm.inc"
 
 #define EXT(a,b) { a , sizeof(a)-1, \
 	extension ## b, NUM(extension ## b), \
-	events ## b, NUM(events ## b)}
+	events ## b, NUM(events ## b), \
+	errors ## b, NUM(errors ## b)}
 struct extension extensions[] = {
+	EXT("MIT-SHM",MITSHM),
+	EXT("RANDR",RANDR),
+	EXT("RENDER",RENDER),
 	EXT("SHAPE",SHAPE),
 	EXT("BIG-REQUESTS",BIGREQUEST)
 };
@@ -1092,3 +1439,4 @@ struct extension *find_extension(u8 *name,size_t len) {
 	}
 	return NULL;
 }
+
